from .sql import backfill_sql, history_table_sql, triggers_sql, table_columns_and_types


def configure_history(db, table):
    """Configure history triggers for a table in a database."""
    # Get the table schema
    columns_and_types = table_columns_and_types(db, table)
    columns = [r[0] for r in columns_and_types]
    with db:
        db.executescript(history_table_sql(table, columns_and_types))
        # Create the triggers
        db.executescript(triggers_sql(table, columns))
        # Backfill the history table
        db.execute(backfill_sql(table, columns))



import re


def history_table_sql(table, columns_and_types):
    """Return SQL for history table for a table and its columns."""
    if isinstance(columns_and_types, dict):
        columns_and_types = dict.items()
    column_names = ",\n".join(
        "   {name} {type}".format(name=escape_sqlite(name), type=type)
        for name, type in columns_and_types
    )
    return """
CREATE TABLE _{table}_history (
    _rowid INTEGER,
{column_names},
    _version INTEGER,
    _updated INTEGER,
    _mask INTEGER
);
CREATE INDEX idx_{table}_history_rowid ON _{table}_history (_rowid);
""".format(
        table=table, column_names=column_names
    )


def triggers_sql(table, columns):
    """Return SQL for triggers for a table and its columns."""
    column_names = ", ".join(escape_sqlite(column) for column in columns)
    new_column_values = ", ".join("new." + escape_sqlite(column) for column in columns)
    old_column_values = ", ".join("old." + escape_sqlite(column) for column in columns)
    # mask is a bit mask of all columns, so len(columns)
    mask = 2 ** len(columns) - 1
    insert_trigger = """
CREATE TRIGGER {table}_insert_history
AFTER INSERT ON {table}
BEGIN
    INSERT INTO _{table}_history (_rowid, {column_names}, _version, _updated, _mask)
    VALUES (new.rowid, {new_column_values}, 1, cast((julianday('now') - 2440587.5) * 86400 * 1000 as integer), {mask});
END;
""".format(
        table=table,
        column_names=column_names,
        new_column_values=new_column_values,
        mask=mask,
    )
    update_columns = []
    for column in columns:
        update_columns.append(
            """
        CASE WHEN old.{column} != new.{column} then new.{column} else null end""".format(
                column=escape_sqlite(column)
            )
        )
    update_columns_sql = ", ".join(update_columns)
    mask_sql = " + ".join(
        """(CASE WHEN old.{column} != new.{column} then {base} else 0 end)""".format(
            column=escape_sqlite(column),
            base=2**idx,
        )
        for idx, column in enumerate(columns)
    )
    where_sql = " or ".join(
        "old.{column} != new.{column}".format(column=escape_sqlite(column))
        for column in columns
    )
    update_trigger = """
CREATE TRIGGER {table}_update_history
AFTER UPDATE ON {table}
FOR EACH ROW
BEGIN
    INSERT INTO _{table}_history (_rowid, {column_names}, _version, _updated, _mask)
    SELECT old.rowid, {update_columns_sql},
        (SELECT MAX(_version) FROM _{table}_history WHERE _rowid = old.rowid) + 1,
        cast((julianday('now') - 2440587.5) * 86400 * 1000 as integer),
        {mask_sql}
    WHERE {where_sql};
END;
""".format(
        table=table,
        column_names=column_names,
        update_columns_sql=update_columns_sql,
        mask_sql=mask_sql,
        where_sql=where_sql,
    )
    delete_trigger = """
CREATE TRIGGER {table}_delete_history
AFTER DELETE ON {table}
BEGIN
    INSERT INTO _{table}_history (_rowid, {column_names}, _version, _updated, _mask)
    VALUES (
        old.rowid,
        {old_column_values},
        (SELECT COALESCE(MAX(_version), 0) from _{table}_history WHERE _rowid = old.rowid) + 1,
        cast((julianday('now') - 2440587.5) * 86400 * 1000 as integer),
        -1
    );
END;
""".format(
        table=table, column_names=column_names, old_column_values=old_column_values
    )
    return insert_trigger + update_trigger + delete_trigger


def backfill_sql(table, columns):
    column_names = ", ".join(escape_sqlite(column) for column in columns)
    sql = """
INSERT INTO _{table}_history (_rowid, {column_names}, _version, _updated, _mask)
SELECT rowid, {column_names}, 1, cast((julianday('now') - 2440587.5) * 86400 * 1000 as integer), {mask}
FROM {table};
""".format(
        table=table, column_names=column_names, mask=2 ** len(columns) - 1
    )
    return sql


def table_columns_and_types(db, table):
    cursor = db.execute(f"PRAGMA table_info([{table}]);")
    columns_and_types = []
    for row in cursor.fetchall():
        columns_and_types.append((row[1], row[2]))
    return columns_and_types


# From https://www.sqlite.org/lang_keywords.html
reserved_words = set(
    (
        "abort action add after all alter analyze and as asc attach autoincrement "
        "before begin between by cascade case cast check collate column commit "
        "conflict constraint create cross current_date current_time "
        "current_timestamp database default deferrable deferred delete desc detach "
        "distinct drop each else end escape except exclusive exists explain fail "
        "for foreign from full glob group having if ignore immediate in index "
        "indexed initially inner insert instead intersect into is isnull join key "
        "left like limit match natural no not notnull null of offset on or order "
        "outer plan pragma primary query raise recursive references regexp reindex "
        "release rename replace restrict right rollback row savepoint select set "
        "table temp temporary then to transaction trigger union unique update using "
        "vacuum values view virtual when where with without"
    ).split()
)

_boring_keyword_re = re.compile(r"^[a-zA-Z_][a-zA-Z0-9_]*$")


def escape_sqlite(s):
    if _boring_keyword_re.match(s) and (s.lower() not in reserved_words):
        return s
    else:
        return f"[{s}]"



import sqlite_utils
from sqlite_history.sql import history_table_sql, triggers_sql
from sqlite_history.cli import run
from unittest.mock import ANY


def test_create_history_table():
    db = sqlite_utils.Database(memory=True)
    db["test"].create({"id": int, "name": str}, pk="id")
    db.executescript(history_table_sql("test", (("id", "integer"), ("name", "text"))))
    assert db["_test_history"].schema == (
        "CREATE TABLE _test_history (\n"
        "    _rowid INTEGER,\n"
        "   id integer,\n"
        "   name text,\n"
        "    _version INTEGER,\n"
        "    _updated INTEGER,\n"
        "    _mask INTEGER\n"
        ")"
    )


def test_triggers():
    db = sqlite_utils.Database(memory=True)
    db["test"].create({"id": int, "name": str, "order": int}, pk="id")
    db.executescript(history_table_sql("test", (("id", "integer"), ("name", "text"))))
    db.executescript(triggers_sql("test", ["id", "name"]))
    db["test"].insert({"id": 1, "name": "Alice"})
    db["test"].insert({"id": 2, "name": "Bob"})
    db["test"].update(1, {"name": "Alice Smith"})
    assert list(db["test"].rows) == [
        {"id": 1, "name": "Alice Smith", "order": None},
        {"id": 2, "name": "Bob", "order": None},
    ]
    assert list(db["_test_history"].rows) == [
        {
            "_rowid": 1,
            "id": 1,
            "name": "Alice",
            "_version": 1,
            "_updated": ANY,
            "_mask": 3,
        },
        {
            "_rowid": 2,
            "id": 2,
            "name": "Bob",
            "_version": 1,
            "_updated": ANY,
            "_mask": 3,
        },
        {
            "_rowid": 1,
            "id": None,
            "name": "Alice Smith",
            "_version": 2,
            "_updated": ANY,
            "_mask": 2,
        },
    ]


def test_triggers_delete():
    db = sqlite_utils.Database(memory=True)
    db["test"].create({"id": int, "name": str}, pk="id")
    db.executescript(history_table_sql("test", (("id", "integer"), ("name", "text"))))
    db.executescript(triggers_sql("test", ["id", "name"]))
    db["test"].insert({"id": 1, "name": "Alice"})
    db["test"].insert({"id": 2, "name": "Bob"})
    db["test"].update(1, {"name": "Alice Smith"})
    db["test"].delete(1)
    assert list(db["test"].rows) == [
        {"id": 2, "name": "Bob"},
    ]
    assert list(db["_test_history"].rows) == [
        {
            "_rowid": 1,
            "id": 1,
            "name": "Alice",
            "_version": 1,
            "_updated": ANY,
            "_mask": 3,
        },
        {
            "_rowid": 2,
            "id": 2,
            "name": "Bob",
            "_version": 1,
            "_updated": ANY,
            "_mask": 3,
        },
        {
            "_rowid": 1,
            "id": None,
            "name": "Alice Smith",
            "_version": 2,
            "_updated": ANY,
            "_mask": 2,
        },
        {
            "_rowid": 1,
            "id": 1,
            "name": "Alice Smith",
            "_version": 3,
            "_updated": ANY,
            "_mask": -1,
        },
    ]


def test_cli(tmpdir):
    test = tmpdir / "test.db"
    db = sqlite_utils.Database(str(test))
    db["test"].insert({"id": 1, "name": "Alice", "order": 1})
    db["test"].enable_fts(("name",), create_triggers=True)
    assert set(db.table_names()) == {
        "test",
        "test_fts",
        "test_fts_docsize",
        "test_fts_idx",
        "test_fts_data",
        "test_fts_config",
    }
    run([str(test), "--all"])
    assert set(db.table_names()) == {
        "test",
        "test_fts",
        "test_fts_docsize",
        "test_fts_idx",
        "test_fts_data",
        "test_fts_config",
        "_test_history",
    }



import argparse
import sqlite3
import os
from . import configure_history


def configure_triggers(database_path, tables):
    db = sqlite3.connect(database_path)
    for table in tables:
        if table.startswith("_") and table.endswith("_history"):
            continue
        # Does a history table exist already?
        history_table_name = f"_{table}_history"
        cursor = db.execute(
            f"SELECT name FROM sqlite_master WHERE type='table' AND name='{history_table_name}';"
        )
        if cursor.fetchone():
            print(f"History table {history_table_name} already exists - skipping.")
            continue
        configure_history(db, table)


def all_regular_tables(database_path):
    """'Regular' excludes FTS and related tables."""
    conn = sqlite3.connect(database_path)
    cursor = conn.cursor()
    hidden_tables = [
        r[0]
        for r in (
            cursor.execute(
                """
            SELECT NAME FROM sqlite_master
            WHERE type = 'table'
            AND (
                sql LIKE '%VIRTUAL TABLE%USING FTS%'
            ) OR name IN ('sqlite_stat1', 'sqlite_stat2', 'sqlite_stat3', 'sqlite_stat4')
        """
            )
        ).fetchall()
    ]
    hidden_tables_copy = hidden_tables[:]
    regular_tables = []
    for row in cursor.execute(
        "SELECT name FROM sqlite_master WHERE type='table';"
    ).fetchall():
        table_name = row[0]
        should_be_hidden = any(
            table_name.startswith(hidden_table) for hidden_table in hidden_tables_copy
        )
        if not should_be_hidden:
            regular_tables.append(table_name)
    return regular_tables


def run(test_args=None):
    parser = argparse.ArgumentParser(
        description="Configure sqlite-history triggers for one or more tables."
    )

    parser.add_argument("database_path", help="Path to the SQLite database file.")
    parser.add_argument(
        "tables", nargs="*", help="One or more table names to configure."
    )
    parser.add_argument(
        "--all", action="store_true", help="Configure for all tables in the database."
    )

    if test_args:
        args = parser.parse_args(test_args)
    else:
        args = parser.parse_args()

    if args.all:
        args.tables = all_regular_tables(args.database_path)
    elif len(args.tables) == 0:
        parser.error(
            "No tables provided. Please provide table names or use --all flag."
        )

    # Error if database_path doesn't exist
    if not os.path.exists(args.database_path):
        parser.error("Database file does not exist.")

    # Error if any of the tables don't exist
    all_table_names = all_regular_tables(args.database_path)
    missing_tables = [table for table in args.tables if table not in all_table_names]
    if missing_tables:
        parser.error("The following tables do not exist: " + ", ".join(missing_tables))

    configure_triggers(args.database_path, args.tables)
